import { bundleMDX } from "@contentlayer/core";
import { markdownToHtml } from "@contentlayer/core";
import {
  casesHandled,
  isNotUndefined,
  omit,
  promiseMap,
  promiseMapPool,
  promiseMapToDict,
  traceAsyncFn
} from "@contentlayer/utils";
import { promises as fs } from "fs";
import { promise as glob } from "glob-promise";
import matter from "gray-matter";
import * as yaml from "js-yaml";
import * as os from "os";
import * as path from "path";
import { match } from "ts-pattern";
const fetchAllDocuments = (async ({
  schemaDef,
  filePathPatternMap,
  contentDirPath,
  flags,
  options,
  previousCache
}) => {
  const documentDefNameWithRelativeFilePathArray = await getDocumentDefNameWithRelativeFilePathArray({
    contentDirPath,
    filePathPatternMap
  });
  const concurrencyLimit = os.cpus().length;
  const documents = await promiseMapPool(documentDefNameWithRelativeFilePathArray, ({ documentDefName, relativeFilePath }) => makeCacheItemFromFilePath({
    relativeFilePath,
    schemaDef,
    documentDefName,
    contentDirPath,
    flags,
    options,
    previousCache
  }), concurrencyLimit).then((_) => _.filter(isNotUndefined));
  const cacheItemsMap = Object.fromEntries(documents.map((_) => [_.document._id, _]));
  return { cacheItemsMap };
})["|>"](traceAsyncFn("@contentlayer/source-local/fetchData:fetchAllDocuments", (_) => omit(_, ["previousCache"])));
const getDocumentDefNameWithRelativeFilePathArray = async ({
  filePathPatternMap,
  contentDirPath
}) => {
  const documentDefNameWithRelativeFilePathsArray = await Promise.all(Object.entries(filePathPatternMap).map(async ([documentDefName, filePathPattern]) => ({
    documentDefName,
    relativeFilePaths: await glob(filePathPattern, { cwd: contentDirPath })
  })));
  const documentDefNameWithRelativeFilePathArray = documentDefNameWithRelativeFilePathsArray.flatMap(({ documentDefName, relativeFilePaths }) => relativeFilePaths.map((relativeFilePath) => ({ relativeFilePath, documentDefName })));
  return documentDefNameWithRelativeFilePathArray;
};
const makeCacheItemFromFilePath = (async ({
  relativeFilePath,
  schemaDef,
  documentDefName,
  contentDirPath,
  flags,
  options,
  previousCache
}) => {
  const fullFilePath = path.join(contentDirPath, relativeFilePath);
  const documentHash = (await fs.stat(fullFilePath)).mtime.getTime().toString();
  if (previousCache && previousCache.cacheItemsMap[relativeFilePath] && previousCache.cacheItemsMap[relativeFilePath].documentHash === documentHash) {
    return previousCache.cacheItemsMap[relativeFilePath];
  }
  const fileContent = await fs.readFile(fullFilePath, "utf-8");
  const filePathExtension = relativeFilePath.toLowerCase().split(".").pop();
  const content = match(filePathExtension).with("md", () => {
    const markdown = matter(fileContent);
    return {
      kind: "markdown",
      data: { ...markdown.data, content: markdown.content }
    };
  }).with("mdx", () => {
    const markdown = matter(fileContent);
    return {
      kind: "mdx",
      data: { ...markdown.data, content: markdown.content }
    };
  }).with("json", () => ({ kind: "json", data: JSON.parse(fileContent) })).when((_) => ["yaml", "yml"].includes(_ ?? ""), () => ({ kind: "yaml", data: yaml.load(fileContent) })).otherwise(() => {
    throw new Error(`Unsupported file extension "${filePathExtension}" for ${relativeFilePath}`);
  });
  const isValid = checkSchema({ content, relativeFilePath, documentDefName, schemaDef, flags });
  if (!isValid) {
    return void 0;
  }
  const documentDef = schemaDef.documentDefMap[documentDefName];
  const document = await makeDocument({ documentDef, rawContent: content, schemaDef, relativeFilePath, options });
  const computedValues = await getComputedValues({ documentDef, doc: document });
  if (computedValues) {
    Object.entries(computedValues).forEach(([fieldName, value]) => {
      document[fieldName] = value;
    });
  }
  return { document, documentHash };
})["|>"](traceAsyncFn("@contentlayer/source-local/fetchData:makeDocumentFromFilePath", (_) => omit(_, ["previousCache"])));
function checkSchema({
  schemaDef,
  content,
  relativeFilePath,
  documentDefName,
  flags
}) {
  const documentDef = schemaDef.documentDefMap[documentDefName];
  if (documentDef === void 0) {
    throw new Error(`No matching document definition found for "${relativeFilePath}"`);
  }
  const existingDataFieldKeys = Object.keys(content.data);
  const requiredFields = documentDef.fieldDefs.filter((_) => _.required);
  const misingRequiredFields = requiredFields.filter((fieldDef) => !existingDataFieldKeys.includes(fieldDef.name));
  if (misingRequiredFields.length > 0) {
    const misingRequiredFieldsStr = misingRequiredFields.map((_, i) => `     ${i + 1}: ` + JSON.stringify(_)).join("\n");
    const message = `Missing required fields (type: "${documentDef.name}") for "${relativeFilePath}".
  Missing fields:
${misingRequiredFieldsStr}
`;
    switch (flags.onMissingOrIncompatibleData) {
      case "skip":
        console.log(`Skipping document. Reason: ${message}`);
        return false;
      case "skip-ignore":
        return false;
      case "fail":
        throw new Error(`Error: ${message}`);
    }
  }
  if (flags.onExtraData === "warn") {
    const schemaFieldNames = documentDef.fieldDefs.map((_) => _.name);
    const extraFieldKeys = existingDataFieldKeys.filter((fieldKey) => !schemaFieldNames.includes(fieldKey));
    if (extraFieldKeys.length > 0) {
      console.log(`Warning: Document (type: "${documentDef.name}") contained fields that are not defined in schema for "${relativeFilePath}".

Extra fields:
${extraFieldKeys.map((key) => `  ${key}: ${JSON.stringify(content.data[key])}`).join("\n")}
`);
    }
  }
  return true;
}
const makeDocument = async ({
  rawContent,
  documentDef,
  schemaDef,
  relativeFilePath,
  options
}) => {
  const docValues = await promiseMapToDict(documentDef.fieldDefs, (fieldDef) => getDataForFieldDef({
    fieldDef,
    rawFieldData: rawContent.data[fieldDef.name],
    schemaDef,
    options
  }), (fieldDef) => fieldDef.name);
  const _raw = {
    sourceFilePath: relativeFilePath,
    sourceFileName: path.basename(relativeFilePath),
    sourceFileDir: path.dirname(relativeFilePath),
    fileType: rawContent.kind,
    flattenedPath: getFlattenedPath(relativeFilePath)
  };
  const doc = {
    _typeName: documentDef.name,
    _id: relativeFilePath,
    _raw,
    ...docValues
  };
  return doc;
};
const getFlattenedPath = (relativeFilePath) => {
  return relativeFilePath.split(".").slice(0, -1).join(".").replace(/\/index$/, "");
};
const makeObject = async ({
  rawObjectData,
  fieldDefs,
  typeName,
  schemaDef,
  options
}) => {
  const objValues = await promiseMapToDict(fieldDefs, (fieldDef) => getDataForFieldDef({
    fieldDef,
    rawFieldData: rawObjectData[fieldDef.name],
    schemaDef,
    options
  }), (fieldDef) => fieldDef.name);
  const obj = { _typeName: typeName, _raw: {}, ...objValues };
  return obj;
};
const getDataForFieldDef = async ({
  fieldDef,
  rawFieldData,
  schemaDef,
  options
}) => {
  if (rawFieldData === void 0) {
    if (fieldDef.required) {
      console.error(`Inconsistent data found: ${fieldDef}`);
    }
    return void 0;
  }
  switch (fieldDef.type) {
    case "object":
      const objectDef = schemaDef.objectDefMap[fieldDef.objectName];
      return makeObject({
        rawObjectData: rawFieldData,
        fieldDefs: objectDef.fieldDefs,
        typeName: objectDef.name,
        schemaDef,
        options
      });
    case "inline_object":
      return makeObject({
        rawObjectData: rawFieldData,
        fieldDefs: fieldDef.fieldDefs,
        typeName: "inline_object",
        schemaDef,
        options
      });
    case "reference":
      return rawFieldData;
    case "polymorphic_list":
    case "list":
      return promiseMap(rawFieldData, (rawItemData) => getDataForListItem({ rawItemData, fieldDef, schemaDef, options }));
    case "date":
      return new Date(rawFieldData);
    case "markdown":
      return {
        raw: rawFieldData,
        html: await markdownToHtml({ mdString: rawFieldData, options: options?.markdown })
      };
    case "mdx":
      return {
        raw: rawFieldData,
        code: await bundleMDX({ mdxString: rawFieldData, options: options?.mdx })
      };
    case "boolean":
    case "string":
    case "number":
    case "json":
    case "slug":
    case "text":
    case "url":
    case "enum":
    case "image":
      return rawFieldData;
    default:
      casesHandled(fieldDef);
  }
};
const getDataForListItem = async ({
  rawItemData,
  fieldDef,
  schemaDef,
  options
}) => {
  if (typeof rawItemData === "string") {
    return rawItemData;
  }
  if (fieldDef.type === "polymorphic_list") {
    const objectTypeName = rawItemData[fieldDef.typeField];
    const objectDef = schemaDef.objectDefMap[objectTypeName];
    if (objectDef === void 0) {
      const valueTypeValues = fieldDef.of.filter((_) => _.type === "object").map((_) => _.objectName).join(", ");
      throw new Error(`Invalid value "${objectTypeName}" for type field "${fieldDef.typeField}" for field "${fieldDef.name}".
Needs to be one of the following values: ${valueTypeValues}`);
    }
    return makeObject({
      rawObjectData: rawItemData,
      fieldDefs: objectDef.fieldDefs,
      typeName: objectDef.name,
      schemaDef,
      options
    });
  }
  switch (fieldDef.of.type) {
    case "object":
      const objectDef = schemaDef.objectDefMap[fieldDef.of.objectName];
      return makeObject({
        rawObjectData: rawItemData,
        fieldDefs: objectDef.fieldDefs,
        typeName: objectDef.name,
        schemaDef,
        options
      });
    case "inline_object":
      return makeObject({
        rawObjectData: rawItemData,
        fieldDefs: fieldDef.of.fieldDefs,
        typeName: "inline_object",
        schemaDef,
        options
      });
    default:
      return rawItemData;
  }
};
const getComputedValues = async ({
  doc,
  documentDef
}) => {
  if (documentDef.computedFields === void 0) {
    return void 0;
  }
  const computedValues = await promiseMapToDict(documentDef.computedFields, (field) => field.resolve(doc), (field) => field.name);
  return computedValues;
};
export {
  fetchAllDocuments,
  getDocumentDefNameWithRelativeFilePathArray,
  makeCacheItemFromFilePath
};
//# sourceMappingURL=fetchData.js.map
