import { getArtifactsDir } from "@contentlayer/core";
import { casesHandled } from "@contentlayer/utils";
import * as chokidar from "chokidar";
import { promises as fs } from "fs";
import * as path from "path";
import { defer, fromEvent, of } from "rxjs";
import { map, mergeMap, startWith, tap } from "rxjs/operators";
import { fetchAllDocuments, getDocumentDefNameWithRelativeFilePathArray, makeCacheItemFromFilePath } from "./fetchData";
import { makeCoreSchema } from "./provideSchema";
export * from "./schema";
export * from "./types";
const fromLocalContent = async (argsOrArgsThunk) => {
  const {
    contentDirPath,
    schema: documentDefs_,
    onMissingOrIncompatibleData = "skip",
    onExtraData = "warn",
    ...options
  } = typeof argsOrArgsThunk === "function" ? await argsOrArgsThunk() : argsOrArgsThunk;
  const documentDefs = (Array.isArray(documentDefs_) ? documentDefs_ : Object.values(documentDefs_)).map((_) => _());
  return {
    type: "local",
    provideSchema: () => makeCoreSchema({ documentDefs }),
    fetchData: ({ watch }) => {
      const filePathPatternMap = documentDefs.reduce((acc, documentDef) => ({ ...acc, [documentDef.name]: documentDef.filePathPattern }), {});
      const flags = { onExtraData, onMissingOrIncompatibleData };
      const schemaDef = makeCoreSchema({ documentDefs });
      const initEvent = { _tag: "init" };
      let cache = void 0;
      const updates$ = watch ? defer(() => fromEvent(chokidar.watch(".", {
        cwd: contentDirPath,
        ignoreInitial: true,
        awaitWriteFinish: { stabilityThreshold: 50, pollInterval: 10 }
      }), "all")).pipe(map(chokidarAllEventToCustomUpdateEvent), tap((e) => (e._tag === "update" || e._tag === "deleted") && console.log(`Watch event "${e._tag}": ${e.relativeFilePath}`)), startWith(initEvent)) : of(initEvent);
      return updates$.pipe(mergeMap(async (event) => {
        if (!cache) {
          cache = await loadPreviousCacheFromDisk({ schemaHash: schemaDef.hash });
        }
        return event;
      }), mergeMap(async (event) => {
        switch (event._tag) {
          case "init":
            return fetchAllDocuments({
              schemaDef,
              filePathPatternMap,
              contentDirPath,
              flags,
              options,
              previousCache: cache
            });
          case "deleted": {
            delete cache.cacheItemsMap[event.relativeFilePath];
            return cache;
          }
          case "update": {
            return updateCacheEntry({
              contentDirPath,
              filePathPatternMap,
              cache,
              event,
              flags,
              schemaDef,
              options
            });
          }
          default:
            casesHandled(event);
        }
      })).pipe(tap(async (cache_) => {
        cache = cache_;
        await writeCacheToDisk({ cache, schemaHash: schemaDef.hash });
      }));
    }
  };
};
const updateCacheEntry = async ({
  contentDirPath,
  filePathPatternMap,
  cache,
  event,
  flags,
  schemaDef,
  options
}) => {
  const documentDefNameWithFilePathArray = await getDocumentDefNameWithRelativeFilePathArray({
    contentDirPath,
    filePathPatternMap
  });
  const documentDefName = documentDefNameWithFilePathArray.find((_) => _.relativeFilePath === event.relativeFilePath)?.documentDefName;
  if (!documentDefName) {
    console.log(`Tried to update cache. No matching document def found called ${documentDefName} for file ${event.relativeFilePath}`);
    return cache;
  }
  const cacheItem = await makeCacheItemFromFilePath({
    contentDirPath,
    documentDefName,
    relativeFilePath: event.relativeFilePath,
    flags,
    schemaDef,
    options,
    previousCache: cache
  });
  if (cacheItem) {
    cache.cacheItemsMap[event.relativeFilePath] = cacheItem;
  }
  return cache;
};
const chokidarAllEventToCustomUpdateEvent = ([eventName, relativeFilePath]) => {
  switch (eventName) {
    case "add":
    case "change":
      return { _tag: "update", relativeFilePath };
    case "unlink":
      return { _tag: "deleted", relativeFilePath };
    case "unlinkDir":
    case "addDir":
      return { _tag: "init" };
    default:
      casesHandled(eventName);
  }
};
const loadPreviousCacheFromDisk = async ({ schemaHash }) => {
  const filePath = path.join(getArtifactsDir(), "cache", `${schemaHash}.json`);
  try {
    const file = await fs.readFile(filePath, "utf8");
    return JSON.parse(file);
  } catch (e) {
    return void 0;
  }
};
const writeCacheToDisk = async ({ cache, schemaHash }) => {
  const cacheDirPath = path.join(getArtifactsDir(), "cache");
  const filePath = path.join(cacheDirPath, `${schemaHash}.json`);
  await fs.mkdir(cacheDirPath, { recursive: true });
  await fs.writeFile(filePath, JSON.stringify(cache), "utf8");
};
export {
  fromLocalContent
};
//# sourceMappingURL=index.js.map
