import { hashObject } from "@contentlayer/core";
import { pick, traceFn, uppercaseFirstChar } from "@contentlayer/utils";
const makeCoreSchema = traceFn("@contentlayer/source-local:makeCoreSchema")((schemaDef) => {
  const coreDocumentDefMap = {};
  const coreObjectDefMap = {};
  for (const documentDef of schemaDef.documentDefs) {
    validateDefName({ defName: documentDef.name });
    const fieldDefs = Object.entries(documentDef.fields).map(fieldDefToCoreFieldDef);
    if ((documentDef.fileType === void 0 || documentDef.fileType === "markdown") && fieldDefs.every((_) => _.name !== "content")) {
      fieldDefs.push({
        type: "markdown",
        name: "content",
        label: "Markdown content",
        description: "Default markdown file content",
        default: void 0,
        const: void 0,
        hidden: void 0,
        required: void 0
      });
    }
    if (documentDef.fileType === "mdx" && fieldDefs.every((_) => _.name !== "content")) {
      fieldDefs.push({
        type: "mdx",
        name: "content",
        label: "MDX content",
        description: "Default MDX file content",
        default: void 0,
        const: void 0,
        hidden: void 0,
        required: void 0
      });
    }
    const computedFields = Object.entries(documentDef.computedFields ?? {}).map(([name, computedField]) => ({ ...pick(computedField, ["description", "resolve", "type"]), name }));
    const coreDocumentDef = {
      _tag: "DocumentDef",
      ...pick(documentDef, ["name", "description", "labelField"]),
      label: documentDef.label ?? documentDef.name,
      isSingleton: documentDef.isSingleton ?? false,
      fieldDefs,
      computedFields
    };
    coreDocumentDefMap[documentDef.name] = coreDocumentDef;
  }
  const objectDefs = collectObjectDefs(schemaDef.documentDefs);
  for (const objectDef of objectDefs) {
    validateDefName({ defName: objectDef.name });
    const coreObjectDef = {
      _tag: "ObjectDef",
      ...pick(objectDef, ["name", "description", "labelField"]),
      label: objectDef.label ?? objectDef.name,
      fieldDefs: Object.entries(objectDef.fields).map(fieldDefToCoreFieldDef)
    };
    coreObjectDefMap[coreObjectDef.name] = coreObjectDef;
  }
  const defs = { documentDefMap: coreDocumentDefMap, objectDefMap: coreObjectDefMap };
  const hash = hashObject(defs);
  return { ...defs, hash };
});
const validateDefName = ({ defName }) => {
  const firstChar = defName.charAt(0);
  if (firstChar.toLowerCase() === firstChar) {
    const improvedDefName = uppercaseFirstChar(defName);
    console.warn(`Warning: A document or object definition name should start with a uppercase letter.
You've provided the name "${defName}" - please consider using "${improvedDefName}" instead.
`);
  }
};
const fieldDefToCoreFieldDef = ([name, fieldDef]) => {
  const baseFields = {
    ...pick(fieldDef, ["type", "default", "description", "required", "const", "hidden"]),
    label: fieldDef.label ?? name,
    name
  };
  switch (fieldDef.type) {
    case "list":
      return { ...baseFields, of: fieldListItemsToCoreFieldListDefItems(fieldDef.of) };
    case "polymorphic_list":
      return {
        ...baseFields,
        typeField: fieldDef.typeField,
        of: fieldDef.of.map(fieldListItemsToCoreFieldListDefItems)
      };
    case "object":
      return { ...baseFields, objectName: fieldDef.object().name };
    case "inline_object":
      const fieldDefs = Object.entries(fieldDef.fields).map(fieldDefToCoreFieldDef);
      return { ...baseFields, fieldDefs };
    case "reference":
      return { ...baseFields, documentName: fieldDef.document().name };
    case "enum":
      return { ...baseFields, options: fieldDef.options };
    default:
      return {
        ...pick(fieldDef, ["type", "default", "description", "label", "required", "const", "hidden"]),
        name
      };
  }
};
const fieldListItemsToCoreFieldListDefItems = (listFieldItem) => {
  switch (listFieldItem.type) {
    case "boolean":
    case "string":
      return pick(listFieldItem, ["labelField", "type"]);
    case "object":
      return {
        type: "object",
        labelField: listFieldItem.labelField,
        objectName: listFieldItem.object().name
      };
    case "enum":
      return {
        type: "enum",
        labelField: listFieldItem.labelField,
        options: listFieldItem.options
      };
    case "inline_object":
      return {
        type: "inline_object",
        labelField: listFieldItem.labelField,
        fieldDefs: Object.entries(listFieldItem.fields).map(fieldDefToCoreFieldDef)
      };
  }
};
function collectObjectDefs(documentDefs) {
  const objectDefMap = {};
  const traverseObjectDef = (objectDef) => {
    if (objectDef.name in objectDefMap) {
      return;
    }
    objectDefMap[objectDef.name] = objectDef;
    Object.values(objectDef.fields).forEach(traverseField);
  };
  const traverseField = (field) => {
    switch (field.type) {
      case "object":
        return traverseObjectDef(field.object());
      case "inline_object":
        return Object.values(field.fields).forEach(traverseField);
      case "polymorphic_list":
        return field.of.forEach(traverseListFieldItem);
      case "list":
        return traverseListFieldItem(field.of);
    }
  };
  const traverseListFieldItem = (listFieldItem) => {
    switch (listFieldItem.type) {
      case "object":
        return traverseObjectDef(listFieldItem.object());
    }
  };
  documentDefs.flatMap((_) => Object.values(_.fields)).forEach(traverseField);
  return Object.values(objectDefMap);
}
export {
  makeCoreSchema
};
//# sourceMappingURL=provideSchema.js.map
