import type { ComputedField } from './computed-field';
export declare type SchemaDef = {
    documentDefs: DocumentDef[];
};
export declare type DocumentFileType = 'markdown' | 'mdx' | 'json' | 'yaml';
/** Top level model type */
export declare type DocumentDef<Name extends string = any> = {
    name: Name;
    /**  */
    label?: string;
    description?: string;
    /** the name of the field that will be used as a title of an object */
    labelField?: string;
    /**
     * In the case of `fileType: "md"` the fields are used for the frontmatter
     * and have an implicit field called `content` with the markdown body.
     */
    fields: Record<string, FieldDef>;
    computedFields?: Record<string, ComputedField<Name>>;
    /** Path is relative to the `contentDirPath` config */
    filePathPattern: string;
    /** Default is `md` */
    fileType?: DocumentFileType;
    isSingleton?: boolean;
};
export declare type FieldDefType = FieldDef['type'];
export declare type FieldDef = ListField | PolymorphicListField | StringField | NumberField | BooleanField | JSONField | SlugField | DateField | MarkdownField | MDXField | TextField | UrlField | ImageField | ObjectField | InlineObjectField | ReferenceField | EnumField;
interface FieldBase {
    /** Field name should contain only alphanumeric characters, underscore and a hyphen [A-Za-z0-9_]. Must start with a letter. Must not end with an underscore or a hyphen. */
    /** Should be short enough as some CMS's have restrictions on its length. Some CMS require label to be unique. */
    label?: string;
    /** Short description to editors how the field is to be used */
    description?: string;
    /**
     * Default: false
     */
    required?: boolean;
    /** IS THIS NEEDED? */
    const?: any;
    /** Users will not be able to edit hidden fields, therefore when hiding a field you should specify the default or const properties to populate these fields when new objects are created. */
    hidden?: boolean;
}
export interface ListField extends FieldBase {
    type: 'list';
    default?: any[];
    of: ListFieldItem;
}
export interface PolymorphicListField extends FieldBase {
    type: 'polymorphic_list';
    default?: any[];
    of: ListFieldItem[];
    /** Field needed to distiguish list data items at run time */
    typeField: string;
}
export declare const isListField: (_: FieldDef) => _ is ListField;
export declare type ListFieldItem = ListFieldItemString | ListFieldItemEnum | ListFieldItemBoolean | ListFieldItemObject | ListFieldItemInlineObject;
declare type BaseListFieldItem = {
    labelField?: string;
};
export declare type ListFieldItemString = BaseListFieldItem & {
    type: 'string';
};
export declare type ListFieldItemEnum = BaseListFieldItem & {
    type: 'enum';
    options: string[];
};
export declare type ListFieldItemBoolean = BaseListFieldItem & {
    type: 'boolean';
};
export declare type ListFieldItemObject = BaseListFieldItem & {
    type: 'object';
    object: Thunk<ObjectDef>;
};
export declare type ListFieldItemInlineObject = BaseListFieldItem & {
    type: 'inline_object';
    fields: Record<string, FieldDef>;
};
export declare const isListFieldItemObject: (_: ListFieldItem) => _ is ListFieldItemObject;
export declare type StringField = FieldBase & {
    type: 'string';
    default?: string;
};
export declare type NumberField = FieldBase & {
    type: 'number';
    default?: number;
};
export declare type BooleanField = FieldBase & {
    type: 'boolean';
    default?: boolean;
};
export declare type JSONField = FieldBase & {
    type: 'json';
    default?: any;
};
export declare type SlugField = FieldBase & {
    type: 'slug';
    default?: string;
};
export declare type DateField = FieldBase & {
    type: 'date';
    default?: string;
};
export declare type MarkdownField = FieldBase & {
    type: 'markdown';
    default?: string;
};
export declare type MDXField = FieldBase & {
    type: 'mdx';
    default?: string;
};
export declare type TextField = FieldBase & {
    type: 'text';
    default?: string;
};
export declare type UrlField = FieldBase & {
    type: 'url';
    default?: string;
};
export declare type ImageField = FieldBase & {
    type: 'image';
    default?: string;
};
export declare type EnumField = FieldBase & {
    type: 'enum';
    default?: any;
    options: string[];
};
export declare type ObjectField = FieldBase & {
    type: 'object';
    default?: any;
    object: Thunk<ObjectDef>;
};
export declare const isObjectField: (_: FieldDef) => _ is ObjectField;
export declare type InlineObjectField = FieldBase & {
    type: 'inline_object';
    default?: any;
    fields: Record<string, FieldDef>;
};
export declare const isInlineObjectField: (_: FieldDef) => _ is InlineObjectField;
export declare type ReferenceField = FieldBase & {
    type: 'reference';
    default?: string;
    document: Thunk<DocumentDef>;
};
export declare type ObjectDef = {
    name: string;
    /**  */
    label?: string;
    description?: string;
    labelField?: string;
    fields: Record<string, FieldDef>;
};
export declare type Thunk<T> = () => T;
export declare const defineObject: (_: () => ObjectDef) => () => ObjectDef;
export declare const defineDocument: <N extends string>(_: () => DocumentDef<N>) => () => DocumentDef<N>;
export declare const defineSchema: (_: SchemaDef) => SchemaDef;
export {};
//# sourceMappingURL=schema.d.ts.map